#include "SoundGen.h"

#include <cmath>
#include <cstdlib>
#include <ctime>
#include <iostream>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif


/* Class Sound */

Sound::Sound () {
    bitDepth = 16;
    sampleRate = 44100.0f;
    noteFreq = 440.0f;
    out = 0.0f;
    outS = 0.0f;
    srand(time(0));
    maxAmplitude = pow(2.0f,bitDepth) / 2.0f;
    triangleCtrl = 1;
}

void Sound::setSound (int bits, float srate, float note) {
    bitDepth = bits;
    sampleRate = srate;
    noteFreq = note;
    maxAmplitude = pow(2.0f,bitDepth) / 2.0f;
}

void Sound::setNote(int midiNote) {
    noteFreq = this->midiToFreq(midiNote);
}

float Sound::sineGen () {
        float increase = (float)((2 * M_PI * noteFreq) / sampleRate);
	
        if (outS < (2 * M_PI))
            outS += increase;
        else
            outS = 0;

        out = maxAmplitude * sin(outS);
	
	return out;
}

float Sound::squareGen () {
        float increase = (float)((2 * M_PI * noteFreq) / sampleRate);
	
        if (outS < (2 * M_PI))
            outS += increase;
        else
            outS = 0;

        out = maxAmplitude * sin(outS);
	
	if (out < 0)
		out = -(maxAmplitude);
	else
		out = maxAmplitude;
	
	return out;  
}

float Sound::sawGen () {
    float increase = (maxAmplitude * (noteFreq / sampleRate));

    if (out < maxAmplitude)
        out += increase;
    else
        out = -(maxAmplitude);

    return out;
}

float Sound::triangleGen () {
    float increase = (2.0f * maxAmplitude * (noteFreq / sampleRate));

    if (triangleCtrl == 1) {
        if (out < maxAmplitude)
            out += increase;
        else
        {
            triangleCtrl = 0;
            out -= increase;
        }
    }
    if (triangleCtrl == 0) {
        if (out > -maxAmplitude)
            out -= increase;
        else
        {
            triangleCtrl = 1;
            out += increase;
        }
    }

    return out;
}

float Sound::noiseGen () {
	union {float f; unsigned int u;} noise;
	unsigned int exp, man, sign;
	sign = rand() % 2;
	exp = rand() % 127;
	man = rand() % 0x800000;
	if (exp > 0) noise.u = (sign << 31) + (exp << 23) + man;
	else noise.u = (sign << 31);
 
	out = (maxAmplitude * noise.f);
  
	return out;
}

float Sound::midiToFreq (int midiNote) {
	if (midiNote <= 127 && midiNote >= 0)
		noteFreq = 440.0f * pow(2.0f,(((float)midiNote-69)/12));
	else
	{
		std::cout << "ERR: MIDI note should be in 0 - 127 range" << std::endl;
		return -1;
	}

	return noteFreq;
}

float Sound::getOut () {
    return out;
}


/* Class SoundFX */

SoundFX::SoundFX () {
	outM = 0;
	outL = 0;
	outR = 0;
	chCheck = 1;
}

float SoundFX::panner (float leftCh, float rightCh, float panValue, short chCheck) {
	outL = leftCh * panValue;
	outR = rightCh * (1.0f - panValue);

	if (chCheck == 1)
		return outL;
	if (chCheck == 2)
		return outR;
	else
	{
		std::cout << "ERR: Channel selection error!" << std::endl;
		return -1;
	}
}

float SoundFX::ampMono (float monoCh, float ampFactor) {
	outM = monoCh * ampFactor;
	
	return outM;
}

float SoundFX::ampStereo (float leftCh, float rightCh, float ampFactor, short chCheck) {
	outL = leftCh * ampFactor;
	outR = rightCh * ampFactor;
	
	if (chCheck == 1)
		return outL;
	if (chCheck == 2)
		return outR;
	else
	{
		std::cout << "ERR: Channel selection error!" << std::endl;
		return -1;
	}
}

float SoundFX::monoToStereo (float monoCh, float leftBal, float rightBal, short chCheck) {
	outL = monoCh * leftBal;
	outR = monoCh * rightBal;
	
	if (chCheck == 1)
		return outL;
	if (chCheck == 2)
		return outR;
	else
	{
		std::cout << "ERR: Channel selection error!" << std::endl;
		return -1;
	}
}

float SoundFX::stereoToMono (float leftCh, float rightCh, float leftChBal, float rightChBal) {
	float totChBal = leftChBal + rightChBal;

	if (totChBal <= 0 || leftChBal * rightChBal <= 0)
	{
		std::cout << "ERR: The two channels' volume values should be positive!" << std::endl;
		return -1;
	}
	
	outM = ((leftCh * leftChBal) / totChBal) + ((rightCh * rightChBal) / totChBal);

	return outM;
}

float SoundFX::cross (float sendCh, float receiveCh, float amount, short chCheck) {
	if (chCheck == 1)
		return outL = receiveCh + (sendCh * amount);
	if (chCheck == 2)
		return outR = receiveCh + (sendCh * amount);
	else
	{
		std::cout << "ERR: Channel selection error!" << std::endl;
		return -1;
	}
}


/* Class ADSR */

ADSR::ADSR () {
        attackSamples = 1;
        decaySamples = 1;
        sustainVol = 1.0f;
        releaseSamples = 1;
	out = 0.0f;
	sampleCount = 0;
	afterCount = 0;
	sampleRate = 44100.0f;
	amplitudeOut = 0.0f;
	noteOn = false;
}

void ADSR::setEnvelope (float sRate, float timeA, float timeD, float volS, float timeR) {
	sampleRate = sRate;
        if (timeA > 0)
            attackSamples = (unsigned long int)((sampleRate / 1000.0f) * timeA);
        else
            attackSamples = 1;
        if (timeD > 0)
            decaySamples = attackSamples + (unsigned long int)((sampleRate / 1000.0f) * timeD);
        else
            decaySamples = attackSamples + 1;
        if (volS >= 0.0f)
            sustainVol = volS;
        else
            sustainVol = 0.0f;
        if (timeR > 0)
            releaseSamples = (unsigned long int)((sampleRate / 1000.0f) * timeR);
        else
            releaseSamples = 1;
}

void ADSR::turnOn () {
	noteOn = true;
	sampleCount = 0;
	amplitudeOut = 0.0f;
}

float ADSR::runEnvelope (float in) {
	if (noteOn) {
                if (sampleCount <= attackSamples) {
			out = in * amplitudeOut;
                        amplitudeOut += (1.0f / attackSamples);
		}
                if (sampleCount > attackSamples && sampleCount <= decaySamples) {
			out = in * amplitudeOut;
                        amplitudeOut -= ((1.0f - sustainVol) / decaySamples);
		}
                if (sampleCount > decaySamples) {
			out = in * amplitudeOut;
		}
	}

	if (!noteOn) {
                if (afterCount <= releaseSamples) {
			out = in * amplitudeOut;
                        amplitudeOut -= (sustainVol / releaseSamples);
		}
		else
			out = 0;
	}
	
	if (noteOn) sampleCount++;
	if (!noteOn) afterCount++;
	
	return out;
}

void ADSR::turnOff () {
	noteOn = false;
	afterCount = 0;
}

void ADSR::resetCount () {
	sampleCount = 0;
	afterCount = 0;
	amplitudeOut = 0.0f;
}

float ADSR::getAttackTime () {
    return (float)((attackSamples / sampleRate) * 1000.0f);
}

float ADSR::getAttackSamples () {
    return (float)attackSamples;
}


float ADSR::getDecayTime () {
    return (float)(((decaySamples / sampleRate) * 1000.0f) - ((attackSamples / sampleRate) * 1000.0f));
}

float ADSR::getDecaySamples() {
    return (float)(decaySamples - attackSamples);
}


float ADSR::getSustain() {
    return sustainVol;
}

float ADSR::getReleaseTime() {
    return (float)((releaseSamples / sampleRate) * 1000.0f);
}

float ADSR::getReleaseSamples() {
    return (float)releaseSamples;
}

void ADSR::setAttackTime(float valueMillisecs) {
    attackSamples = (valueMillisecs * sampleRate) / 1000.0f;
}

void ADSR::setDecayTime(float valueMillisecs) {
    decaySamples = (valueMillisecs * sampleRate) / 1000.0f;
}

void ADSR::setSustain(float value) {
    sustainVol = value;
}

void ADSR::setReleaseTime(float valueMillisecs) {
    releaseSamples = (valueMillisecs * sampleRate) / 1000.0f;
}
